/*
 * Affiche d'un fichier .ppm
 *
 * Auteur: Pierre Ficheux (pierre@alienor.fr)  Novembre 1999
 */
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

GC	gc;
Display	*display;
int	screen, nplanes;
Window	win, root;
unsigned long white_pixel, black_pixel;
static Colormap current_cmap, default_cmap;
static Visual *the_visual;

static int red_bits;
static int green_bits;
static int blue_bits;
static int red_shift;
static int green_shift;
static int blue_shift;
static unsigned long red_mask;
static unsigned long green_mask;
static unsigned long blue_mask;

static char magic[3];
static int nb_colors;

static int last_color;
static unsigned short red, green, blue;

struct {
  unsigned short red, green, blue;
  unsigned long pixel;
} color_cache[256];


FILE *fp;
XImage *ximage;
char *image_data;
int image_width = 100;
int image_height = 100;

/*
 * fonction associee a l'evenement EXPOSE 
 */
void expose ()
{
    XPutImage (display, win, gc, ximage, 0, 0, 0, 0, image_width, image_height);
}

/*
 * programme principal
 */

main(ac, av)
int ac;
char **av;
{
    unsigned long r, g, b;
    register int x, y, i;
    long pixel;
    XSetWindowAttributes attr;

    if (ac == 1) {
      fprintf (stderr, "Usage: xppm ppm_file\n");
      exit (1);
    }

    if (!(fp = fopen (av[1], "r"))) {
	perror (av[1]);
	exit (1);
    }

    if ((display = XOpenDisplay (NULL)) == NULL) {
	fprintf (stderr, "Can't open Display\n");
	exit (1);
    }

    gc = DefaultGC (display, screen);
    screen = DefaultScreen (display);
    root = RootWindow (display, screen);
    white_pixel = WhitePixel (display, screen);
    black_pixel = BlackPixel (display, screen);
    default_cmap = DefaultColormap (display, screen);
    nplanes = DisplayPlanes (display, screen);
    the_visual = DefaultVisual (display, screen);

    if (nplanes > 8) {
      current_cmap = default_cmap;

      /* Calcul des decalages */
      r = red_mask = the_visual->red_mask;
      g = green_mask = the_visual->green_mask;
      b = blue_mask = the_visual->blue_mask;
	
      red_bits = 0;
      green_bits = 0;
      blue_bits = 0;
      red_shift = 0;
      green_shift = 0;
      blue_shift = 0;
    
      while (!(r & 1)) {
	r >>= 1;
	red_shift++;
      }
      while (r & 1) {
	r >>= 1;
	red_bits++;
      }
      while (!(g & 1)) {
	g >>= 1;
	green_shift++;
      }
      while (g & 1) {
	g >>= 1;
	green_bits++;
      }
      while (!(b & 1)) {
	b >>= 1;
	blue_shift++;
      }
      while (b & 1) {
	b >>= 1;
	blue_bits++;
      }
    }
    else
      current_cmap = XCopyColormapAndFree (display, default_cmap);

    /* Lecture de l'en-tte: ATTENTION, on ne traite pas les commentaires ! */
    fscanf (fp, "%c%c\n", &magic[0], &magic[1]);

    if (magic[1] != '6') {
      fprintf (stderr, "%s: %s is not a PPM file.\n", av[0], av[1]);
      fclose (fp);
      exit (1);
    }

    fscanf (fp, "%d %d\n", &image_width, &image_height);
    fscanf (fp, "%d\n", &nb_colors);

    /* Allocation des datas */
    if (!(image_data = (char*) calloc (1, sizeof(unsigned int) * image_width * image_height))) {
	perror ("calloc image_data:");
	exit (1);
    }

    /* Cration de l'image */
    ximage = XCreateImage(display, DefaultVisual (display, screen), nplanes, ZPixmap, 0, image_data, image_width, image_height, 8 * sizeof(unsigned int), 0) ;
#ifdef __i386__		
    ximage->byte_order = LSBFirst ;
#else
    ximage->byte_order = MSBFirst ;
#endif

    for (y = 0 ; y < image_height ; y++) {
      for (x = 0 ; x < image_width ; x++) {

	fscanf (fp, "%c%c%c", &red, &green, &blue);

	if (nplanes > 8) {
	  red >>= (8 - red_bits);
	  green >>= (8 - green_bits);
	  blue >>= (8 - blue_bits);

	  pixel = ((red << red_shift) & red_mask) | ((green << green_shift) & green_mask) | ((blue << blue_shift) & blue_mask);

	}
	/* 8 planes */
	else {
	  XColor color;

	  red <<= 8;
	  green <<= 8;
	  blue <<= 8;

	  for (i = 0 ; i != last_color ; i++) {
	    if (color_cache[i].red == red && color_cache[i].green == green && color_cache[i].blue == blue)
	      break;
	  }
	    
	  if (i == last_color) {
	    color.red = red;
	    color.green = green;
	    color.blue = blue;

	    if (!XAllocColor (display, current_cmap, &color)) {
	      fprintf (stderr, "Can't allocate %d %d %d\n", red, green, blue);
	      exit (1);
	    }
	    else {
	      color_cache[i].red = red;
	      color_cache[i].green = green;
	      color_cache[i].blue = blue;
	      color_cache[i].pixel = pixel = color.pixel;
	    }

	    last_color++;
	    if (last_color > 256)  {
	      fprintf (stderr, "Too many colors...\n");
	      exit (1);
	    }
	  }
	  else
	    pixel = color_cache[i].pixel;
	}

	XPutPixel (ximage, x, y, pixel);
      }
    }

    fclose (fp);

    attr.background_pixel = white_pixel;
    attr.border_pixel = black_pixel;
    attr.colormap = current_cmap;

    win = XCreateWindow (display, root, 0, 0, image_width, image_height, 2, nplanes, CopyFromParent, the_visual, CWBackPixel | CWBorderPixel | CWColormap, &attr);

    XSelectInput (display, win , ExposureMask | ButtonPressMask);

    XMapWindow (display, win);

    for (;;) {
	XEvent ev;

	XNextEvent (display, &ev);

	switch (ev.type) {
	  case Expose :
	    expose ();
	    break;

	  case ButtonPress :
	    exit (0);
	    break;

	    default : break;
	}
    }
}
